import functools
from typing import List
import jax
import jax.numpy as jnp

from ttax.base_class import TT
from ttax.base_class import TTMatrix
from ttax.base_class import TTTensOrMat
from ttax import compile
from ttax.ops import tt_vmap
from ttax.decompositions import orthogonalize


@tt_vmap()
def tangent_to_deltas(tangent_element: TTTensOrMat) -> List[jnp.ndarray]:
  """Convert an element of the tangent space to deltas representation.
  Tangent space elements (outputs of ``ttax.project``) look like:
  ``dP1 V2 ... Vd + U1 dP2 V3 ... Vd + ... + U1 ... Ud-1 dPd``.
  This function takes as input an element of the tangent space and converts
  it to the list of deltas: ``[dP1, ..., dPd]``.
  
  :type tangent_element: `TT-Tensor` or `TT-Matrix` 
  :param tangent_element: a result of ``ttax.project``
  :return: list of delta-cores
  :rtype: list
  """
  # TODO: project on
  ndim = tangent_element.ndim
  tt_ranks = tangent_element.tt_ranks
  deltas = [None] * ndim
  for i in range(1, ndim - 1):
    if int(tt_ranks[i] / 2) != float(tt_ranks[i]) / 2:
      raise ValueError('tangent_element argument is supposed to be a '
                       'projection, but its ranks are not even.')
  for i in range(1, ndim - 1):
    r1, r2 = tt_ranks[i], tt_ranks[i + 1]
    curr_core = tangent_element.tt_cores[i]
    deltas[i] = curr_core[int(r1 / 2):, ..., :int(r2 / 2)]
  deltas[0] = tangent_element.tt_cores[0][..., :int(tt_ranks[1] / 2)]
  deltas[ndim - 1] = tangent_element.tt_cores[ndim - 1][int(tt_ranks[-2] / 2):]
  return deltas


def _deltas_tt_vmap(_deltas_to_tangent):
  """Like tt_vmap, but tailored to be used with deltas_to_tangent."""
  @functools.wraps(_deltas_to_tangent)
  def vectorized_deltas_to_tangent(deltas, tt):
    if tt.is_tt_matrix:
      num_batch_dims = len(deltas[0].shape) - 4
    else:
      num_batch_dims = len(deltas[0].shape) - 3
    if num_batch_dims == 0:
      return _deltas_to_tangent(deltas, tt)
    else:
      # Vmap everything num_batch_dims times.
      vmapped = _deltas_to_tangent
      for _ in range(num_batch_dims):
        vmapped = jax.vmap(vmapped, in_axes=(0, 0))
      return vmapped(deltas, tt)
  return vectorized_deltas_to_tangent


@_deltas_tt_vmap
def deltas_to_tangent(deltas: List[jnp.ndarray],
                      tt: TTTensOrMat) -> TTTensOrMat:
  """Converts deltas representation of tangent space vector to `TT-object`.
  Takes as input a list of [dP1, ..., dPd] and returns
  dP1 V2 ... Vd + U1 dP2 V3 ... Vd + ... + U1 ... Ud-1 dPd.
  
  This function is hard to use correctly because deltas should obey the
  so called gauge conditions. If they don't, the function will silently return
  incorrect result. That is why this function is not imported in __init__.
  
  :param deltas: a list of deltas (essentially `TT-cores`) obeying the gauge
                 conditions.
  :param tt: object on which the tangent space tensor represented
             by delta is projected.
  :type tt: `TT-Tensor` or `TT-Matrix`
  :return: object constructed from deltas, that is from the tangent
           space at point `tt`.
  :rtype: `TT-Tensor` or `TT-Matrix`
  """
  cores = []
  dtype = tt.dtype
  left = orthogonalize(tt)
  right = orthogonalize(left, left_to_right=False)
  left_rank_dim = 0
  right_rank_dim = 3 if tt.is_tt_matrix else 2
  for i in range(tt.ndim):
    left_tt_core = left.tt_cores[i]
    right_tt_core = right.tt_cores[i]

    if i == 0:
      tangent_core = jnp.concatenate((deltas[i], left_tt_core),
                                     axis=right_rank_dim)
    elif i == tt.ndim - 1:
      tangent_core = jnp.concatenate((right_tt_core, deltas[i]),
                                     axis=left_rank_dim)
    else:
      rank_1 = right.tt_ranks[i]
      rank_2 = left.tt_ranks[i + 1]
      if tt.is_tt_matrix:
        mode_size_n = tt.raw_tensor_shape[0][i]
        mode_size_m = tt.raw_tensor_shape[1][i]
        shape = [rank_1, mode_size_n, mode_size_m, rank_2]
      else:
        mode_size_n = tt.shape[i]
        shape = [rank_1, mode_size_n, rank_2]
      zeros = jnp.zeros(shape, dtype=dtype)
      upper = jnp.concatenate((right_tt_core, zeros), axis=right_rank_dim)
      lower = jnp.concatenate((deltas[i], left_tt_core), axis=right_rank_dim)
      tangent_core = jnp.concatenate((upper, lower), axis=left_rank_dim)
    cores.append(tangent_core)
  if tt.is_tt_matrix:
    return TTMatrix(cores)
  else:
    return TT(cores)



import jax
import jax.numpy as jnp

from typing import Callable

from ttax import decompositions
from ttax import riemannian
from ttax import ops

from ttax.base_class import TTTensOrMat

TangentVector = TTTensOrMat


def _enforce_gauge_conditions(deltas, left):
  """Project deltas that define tangent space vec onto the gauge conditions."""
  proj_deltas = []
  for i in range(left.ndim):
    right_r = left.tt_ranks[i + 1]
    q = left.tt_cores[i].reshape((-1, right_r))
    if i < left.ndim - 1:
      proj_delta = deltas[i]
      proj_delta = proj_delta.reshape((-1, right_r))
      proj_delta -= q @ (q.T @ proj_delta)
      proj_delta = proj_delta.reshape(left.tt_cores[i].shape)
    else:
      proj_delta = deltas[i]
    proj_deltas.append(proj_delta)
  return proj_deltas


def grad(func: Callable[[TTTensOrMat], float]) -> Callable[[TTTensOrMat], TangentVector]:
  """Riemannian autodiff: decorator to compute gradient projected on tangent space.

  Returns a function which at a point X computes projection of the euclidian gradient
  df/dx onto the tangent space of TT-tensors at point x.
  Warning: grad may not work for some function, e.g. ones that include QR or
  SVD decomposition (ttax.project, ttax.round) or for functions that work
  with TT-cores directly (in contrast to working with TT-object only via ttax
  functions). In this cases this function can silently return wrong results!

  Example:
      # Scalar product with some predefined tensor squared 0.5 * <x, t>**2.
      # It's gradient is <x, t> t and it's Riemannian gradient is
      #     ttax.project(<x, t> * t, x)
      f = lambda x: 0.5 * ttax.flat_inner(x, t)**2
      # Equivalent to
      #   projected_grad = ttax.project(ttax.flat_inner(x, t) * t, x)
      projected_grad = ttax.grad(f)(x)

  Args:
      func: function that takes TensorTrain object as input and outputs a number.

  Returns:
      Function that computes Riemannian gradient of `func` at a given point.

  See also:
      ttax.hessian_vector_product
  """
  def _grad(x: TTTensOrMat) -> TangentVector:
    # TODO: support runtime checks
    left = decompositions.orthogonalize(x)
    right = decompositions.orthogonalize(left, left_to_right=False)
    deltas = [right.tt_cores[0]]
    deltas += [jnp.zeros_like(cc) for cc in right.tt_cores[1:]]

    def augmented_func(d):
      x_projection = riemannian.deltas_to_tangent(d, x)
      return func(x_projection)

    function_value, cores_grad = jax.value_and_grad(augmented_func)(deltas)

    deltas = _enforce_gauge_conditions(cores_grad, left)
    return riemannian.deltas_to_tangent(deltas, x)
  return _grad


def hessian_vector_product(func: Callable[[TTTensOrMat], float]) -> Callable[[TTTensOrMat, TangentVector], TangentVector]:
  """P_x [d^2f/dx^2] P_x vector, i.e. Riemannian hessian by vector product.

    Computes
      P_x [d^2f/dx^2] P_x vector
    where P_x is projection onto the tangent space of TT at point x and
    d^2f/dx^2 is the Hessian of the function.
    Note that the true Riemannian hessian also includes the manifold curvature
    term which is ignored here.
    Warning: this is experimental feature and it may not work for some function,
    e.g. ones that include QR or SVD decomposition (t3f.project, t3f.round) or
    for functions that work with TT-cores directly (in contrast to working with
    TT-object only via t3f functions). In this cases this function can silently
    return wrong results!

    Example:
        # Quadratic form with matrix A: <x, A x>.
        # It's gradient is (A + A.T) x, it's Hessian is (A + A.T)
        # It's Riemannian Hessian by vector product is
        #     proj_vec = t3f.project(vector, x)
        #     t3f.project(t3f.matmul(A + t3f.transpose(A), proj_vec), x)
        f = lambda x: t3f.bilinear_form(A, x, x)
        res = t3f.hessian_vector_product(f, x, vector)

    Args:
        func: function that takes TensorTrain object as input and outputs a number.

    Returns:
        `TensorTrain`, result of the Riemannian hessian by vector product.

    See also:
        `ttax.grad`
    """
  def _hess_by_vec(x: TTTensOrMat, vector: TangentVector) -> TangentVector:
    left = decompositions.orthogonalize(x)
    right = decompositions.orthogonalize(left, left_to_right=False)
    deltas = [right.tt_cores[0]]
    deltas += [jnp.zeros_like(cc) for cc in right.tt_cores[1:]]

    def augmented_outer_func(deltas_outer):

      def augmented_inner_func(deltas_inner):
        x_projection = riemannian.deltas_to_tangent(deltas_inner, x)
        return func(x_projection)

      function_value, cores_grad = jax.value_and_grad(augmented_inner_func,
                                                      deltas_outer)
      # TODO: support runtime checks

      vector_projected = riemannian.project(vector, x)
      vec_deltas = riemannian.tangent_space_to_deltas(vector_projected)
      products = [jnp.sum(a * b) for a, b in zip(cores_grad, vec_deltas)]
      return sum(products)

    _, second_cores_grad = jax.value_and_grad(augmented_outer_func, deltas)
    final_deltas = _enforce_gauge_conditions(second_cores_grad, left)
    # TODO: pass left and right?
    return riemannian.deltas_to_tangent(final_deltas, x)
  return _hess_by_vec


def project(what, where):
  """Project `what` TTs on the tangent space of `where` TT.

  ``project(what, x) = P_x(what)``
  ``project(batch_what, x) = batch(P_x(batch_what[0]), ..., P_x(batch_what[N]))``

  Complexity:
    ``O(d r_where^3 m)`` for orthogonalizing the TT-cores of `where`
    ``+O(batch_size d r_what r_where n (r_what + r_where))``

      - ``d`` is the number of `TT-cores`: ``what.ndims()``
      - ``r_what`` is the largest `TT-rank` of `what`: ``max(what.tt_rank())``
      - ``r_where`` is the largest `TT-rank` of `where`
      - ``n`` is the size of the axis dimension of `what` and `where` e.g.
        for a tensor of size 4 x 4 x 4, ``n`` is 4;
        for a 9 x 64 matrix of raw shape (3, 3, 3) x (4, 4, 4) ``n`` is 12

  :type what: `TT-Tensor` or `TT-Matrix`
  :param what: in the case of batch returns batch with projection of each individual tensor
  :type where: `TT-Tensor` or `TT-Matrix`
  :param where: on which tangent space to project
  :return: `TT-object` with the `TT-ranks` equal ``2 * tangent_space_tens.tt_ranks()``
  :rtype: `TT-Tensor` or `TT-Matrix`
  """
  def _f(x):
    return ops.flat_inner(what, x)
  return grad(_f)(where)


from absl.testing import absltest

import numpy as np
import jax
import jax.numpy as jnp
import jax.test_util as jtu
from jax.config import config

from ttax import random_
from ttax import decompositions
from ttax import ops
from ttax.base_class import TT

config.parse_flags_with_absl()


class DecompositionsTest(jtu.JaxTestCase):

  def testOrthogonalizeLeftToRight(self):
    dtype = jnp.float32
    rng = jax.random.PRNGKey(0)
    shape = (2, 4, 3, 3)
    tt_ranks = (1, 5, 2, 17, 1)
    updated_tt_ranks = (1, 2, 2, 6, 1)
    tens = random_.tensor(rng, shape, tt_rank=tt_ranks, dtype=dtype)
    orthogonal = decompositions.orthogonalize(tens)

    self.assertAllClose(ops.full(tens), ops.full(orthogonal), atol=1e-5,
                        rtol=1e-5)
    self.assertArraysEqual(updated_tt_ranks, orthogonal.tt_ranks)
    # Check that the TT-cores are orthogonal.
    for core_idx in range(4 - 1):
      core = orthogonal.tt_cores[core_idx]
      core = jnp.reshape(core, (updated_tt_ranks[core_idx] * shape[core_idx],
                                updated_tt_ranks[core_idx + 1]))
      should_be_eye = core.T @ core
      self.assertAllClose(np.eye(updated_tt_ranks[core_idx + 1]), should_be_eye)

  def testOrthogonalizeRightToLeft(self):
    dtype = jnp.float32
    rng = jax.random.PRNGKey(0)
    shape = (2, 4, 3, 3)
    tt_ranks = (1, 5, 2, 17, 1)
    updated_tt_ranks = (1, 5, 2, 3, 1)
    tens = random_.tensor(rng, shape, tt_rank=tt_ranks, dtype=dtype)
    orthogonal = decompositions.orthogonalize(tens, left_to_right=False)

    self.assertAllClose(ops.full(tens), ops.full(orthogonal), atol=1e-5,
                        rtol=1e-5)
    self.assertArraysEqual(updated_tt_ranks, orthogonal.tt_ranks)
    # Check that the TT-cores are orthogonal.
    for core_idx in range(1, 4):
      core = orthogonal.tt_cores[core_idx]
      core = jnp.reshape(core, (updated_tt_ranks[core_idx], shape[core_idx] *
                                updated_tt_ranks[core_idx + 1]))
      should_be_eye = core @ core.T
      self.assertAllClose(np.eye(updated_tt_ranks[core_idx]), should_be_eye)

  def testRound2d(self):
    dtype = jnp.float32
    rank = 5
    np.random.seed(0)
    x = np.random.randn(10, 20).astype(dtype)
    u, s, v = np.linalg.svd(x, full_matrices=False)
    core_1 = u @ np.diag(s)
    core_1 = core_1.reshape(1, 10, 10)
    core_2 = v
    core_2 = core_2.reshape(10, 20, 1)
    tt = TT((core_1, core_2))
    truncated_x = u[:, :rank] @ np.diag(s[:rank]) @ v[:rank, :]
    rounded = decompositions.round(tt, 5)
    self.assertAllClose(truncated_x, ops.full(rounded), rtol=1e-5, atol=1e-5)


if __name__ == '__main__':
  absltest.main(testLoader=jtu.JaxTestLoader())



from absl.testing import absltest
from absl.testing import parameterized

import numpy as np
import jax
import jax.numpy as jnp
import jax.test_util as jtu
from jax.config import config

from ttax.base_class import TT
from ttax.base_class import TTMatrix
from ttax import random_
from ttax import ops

config.parse_flags_with_absl()


class TTTensorTest(jtu.JaxTestCase):

  def testFullTensor2d(self):
    np.random.seed(1)
    for rank in [1, 2]:
      a = np.random.rand(10, rank)
      b = np.random.rand(rank, 9)
      tt_cores = (a.reshape(1, 10, rank), b.reshape(rank, 9, 1))
      desired = np.dot(a, b)
      tt_tens = TT(tt_cores)
      actual = ops.full(tt_tens)
      self.assertAllClose(desired, actual)

  def testFullTensor2dBatch(self):
    np.random.seed(1)
    for rank in [1, 2]:
      a = np.random.rand(3, 10, rank)
      b = np.random.rand(3, rank, 9)
      tt_cores = (a.reshape(3, 1, 10, rank), b.reshape(3, rank, 9, 1))
      desired = np.einsum('bij,bjk->bik', a, b)
      tt_tens = TT(tt_cores)
      actual = ops.full(tt_tens)
      self.assertAllClose(desired, actual)

  def testMultiply(self):
    # Multiply two TT-tensors.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (1, 2, 3, 4), tt_rank=2, dtype=dtype)
    tt_b = random_.tensor(rng2, (1, 2, 3, 4), tt_rank=[1, 1, 4, 3, 1],
                          dtype=dtype)

    res_actual1 = ops.full(ops.multiply(tt_a, tt_b))
    res_actual2 = ops.full(tt_a * tt_b)
    res_desired = ops.full(tt_a) * ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired)
    self.assertAllClose(res_actual2, res_desired)

  def testMultiplyBatch(self):
    # Multiply two batches of TT-tensors.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (1, 2, 3, 4), tt_rank=2, batch_shape=(3,),
                          dtype=dtype)
    tt_b = random_.tensor(rng2, (1, 2, 3, 4), tt_rank=[1, 1, 4, 3, 1],
                          batch_shape=(3,), dtype=dtype)

    res_actual1 = ops.full(ops.multiply(tt_a, tt_b))
    res_actual2 = ops.full(tt_a * tt_b)
    res_desired = ops.full(tt_a) * ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired)
    self.assertAllClose(res_actual2, res_desired)

  def testFlatInner(self):
    # Multiply two TT-tensors.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (1, 2, 3, 4), tt_rank=2, dtype=dtype)
    tt_b = random_.tensor(rng2, (1, 2, 3, 4), tt_rank=[1, 1, 4, 3, 1], dtype=dtype)
    res_actual = ops.flat_inner(tt_a, tt_b)
    res_desired = jnp.sum(ops.full(tt_a) * ops.full(tt_b))
    self.assertAllClose(res_actual, res_desired)
    
  def testAdd(self):
    # Add two TT-tensors.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (2, 1, 3, 4), tt_rank=2, dtype=dtype)
    tt_b = random_.tensor(rng2, (2, 1, 3, 4), tt_rank=[1, 2, 4, 3, 1],
                          dtype=dtype)

    res_actual1 = ops.full(ops.add(tt_a, tt_b))
    res_actual2 = ops.full(tt_a + tt_b)
    res_desired = ops.full(tt_a) + ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired)
    self.assertAllClose(res_actual2, res_desired)

  def testAddSameBatchSize(self):
    # Add two batches of TT-tensors.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (2, 1, 3, 4), tt_rank=2, batch_shape=(3,),
                          dtype=dtype)
    tt_b = random_.tensor(rng2, (2, 1, 3, 4), tt_rank=[1, 2, 4, 3, 1],
                          batch_shape=(3,), dtype=dtype)

    res_actual1 = ops.full(ops.add(tt_a, tt_b))
    res_actual2 = ops.full(tt_a + tt_b)
    res_desired = ops.full(tt_a) + ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired)
    self.assertAllClose(res_actual2, res_desired)
    
  def testAddBroadcasting(self):
    # Sum two TT-tensors with broadcasting.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (2, 1, 4), tt_rank=2, 
                          batch_shape=(1,), dtype=dtype)
    tt_b = random_.tensor(rng2, (2, 1, 4), tt_rank=[1, 2, 4, 1],
                          batch_shape=(3,), dtype=dtype)
    
    res_actual1 = ops.full(ops.add(tt_a, tt_b))
    res_actual2 = ops.full(tt_b + tt_a)
    res_desired = ops.full(tt_a) + ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired)
    self.assertAllClose(res_actual2, res_desired)

  def testMultiplyByScalar(self):
    # Multiply batch of TT-tensor by scalar.
    c = 4.5
    rng = jax.random.PRNGKey(0)
    dtype = jnp.float32
    tt = random_.tensor(rng, (2, 1, 3, 4), tt_rank=[1, 2, 4, 3, 1],
                        dtype=dtype)
    res_actual1 = ops.full(ops.multiply(tt, c))
    res_actual2 = ops.full(tt * c)
    res_actual3 = ops.full(c * tt)
    res_desired = c * ops.full(tt)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-4)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-4)
    self.assertAllClose(res_actual3, res_desired, rtol=1e-4)

  def testMultiplyBatchByScalar(self):
    # Multiply batch of TT-tensor by scalar.
    c = 4.5
    rng = jax.random.PRNGKey(0)
    dtype = jnp.float32
    tt = random_.tensor(rng, (2, 1, 3, 4), tt_rank=[1, 2, 4, 3, 1],
                        batch_shape=(3,), dtype=dtype)
    res_actual1 = ops.full(ops.multiply(tt, c))
    res_actual2 = ops.full(tt * c)
    res_actual3 = ops.full(c * tt)
    res_desired = c * ops.full(tt)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-4)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-4)
    self.assertAllClose(res_actual3, res_desired, rtol=1e-4)


class TTMatrixTest(jtu.JaxTestCase):

  def testFull2d(self):
    np.random.seed(1)
    for rank in [1, 2]:
      a = np.random.rand(9, rank)
      b = np.random.rand(rank, 10)
      tt_cores = (a.reshape(1, 3, 3, rank), b.reshape(rank, 2, 5, 1))
      desired = np.einsum('aijb,bpqc->ipjq', *tt_cores)
      desired = desired.reshape(6, 15)
      tt_tens = TTMatrix(tt_cores)
      actual = ops.full(tt_tens)
      self.assertAllClose(desired, actual)

  def testFull2dBatch(self):
    np.random.seed(1)
    for rank in [1, 2]:
      a = np.random.rand(7, 9, rank)
      b = np.random.rand(7, rank, 10)
      tt_cores = (a.reshape(7, 1, 3, 3, rank), b.reshape(7, rank, 2, 5, 1))
      desired = np.einsum('taijb,tbpqc->tipjq', *tt_cores)
      desired = desired.reshape(7, 6, 15)
      tt_tens = TTMatrix(tt_cores)
      actual = ops.full(tt_tens)
      self.assertAllClose(desired, actual)

  def testMatmul(self):
    # Multiply two TT-matrices.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    sum_shape = (4, 3, 5)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, sum_shape), tt_rank=3, dtype=dtype)
    tt_b = random_.matrix(rng2, (sum_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          dtype=dtype)

    res_actual = ops.full(ops.matmul(tt_a, tt_b))
    res_desired = ops.full(tt_a) @ ops.full(tt_b)
    # TODO: why such low precision?
    self.assertAllClose(res_actual, res_desired, rtol=1e-3)

  def testMultiply(self):
    # Elementwise multiply two TT-matrices.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, right_shape), tt_rank=3,
                          dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          dtype=dtype)

    res_actual1 = ops.full(ops.multiply(tt_a, tt_b))
    res_actual2 = ops.full(tt_a * tt_b)
    res_desired = ops.full(tt_a) * ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-4)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-4)

  def testMultiplyBatch(self):
    # Elementwise multiply two batches of TT-matrices.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, right_shape), tt_rank=3,
                          batch_shape=(3,), dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          batch_shape=(3,), dtype=dtype)

    res_actual1 = ops.full(ops.multiply(tt_a, tt_b))
    res_actual2 = ops.full(tt_a * tt_b)
    res_desired = ops.full(tt_a) * ops.full(tt_b)
    # TODO: why such low precision?
    self.assertAllClose(res_actual1, res_desired, rtol=1e-3)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-3)

  def testFlatInner(self):
    # Multiply two TT-matrices.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, right_shape), tt_rank=3,
                          dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          dtype=dtype)
    res_actual = ops.flat_inner(tt_a, tt_b)
    res_desired = jnp.sum(ops.full(tt_a) * ops.full(tt_b))
    self.assertAllClose(res_actual, res_desired)
    
  def testAdd(self):
    # Add two TT-matrices.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, right_shape), tt_rank=3,
                          dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          dtype=dtype)

    res_actual1 = ops.full(ops.add(tt_a, tt_b))
    res_actual2 = ops.full(tt_a + tt_b)
    res_desired = ops.full(tt_a) + ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-5)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-5)

  def testAddSameBatchSize(self):
    # Add two batches of TT-matrices.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, right_shape), tt_rank=3,
                          batch_shape=(3,), dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          batch_shape=(3,), dtype=dtype)

    res_actual1 = ops.full(ops.add(tt_a, tt_b))
    res_actual2 = ops.full(tt_a + tt_b)
    res_desired = ops.full(tt_a) + ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-3)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-3)  
    
  def testAddBroadcasting(self):
    # Sum two TT-Matrices with broadcasting.
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, right_shape), tt_rank=3,
                          batch_shape=(3, 1, 3,), dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, right_shape), tt_rank=[1, 4, 3, 1],
                          batch_shape=(3, 3, 3), dtype=dtype)
    
    res_actual1 = ops.full(ops.add(tt_a, tt_b))
    res_actual2 = ops.full(tt_b + tt_a)
    res_desired = ops.full(tt_a) + ops.full(tt_b)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-4)
    self.assertAllClose(res_actual2, res_desired, rtol=1e-4)

  def testMultiplyByScalar(self):
    # Multiply TT-matrix by scalar.
    c = 4.5
    rng = jax.random.PRNGKey(0)
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt = random_.matrix(rng, (left_shape, right_shape), tt_rank=3,
                        dtype=dtype)
    res_actual1 = ops.full(ops.multiply(tt, c))
    res_actual2 = ops.full(tt * c)
    res_actual3 = ops.full(c * tt)
    res_desired = c * ops.full(tt)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-4)  
    self.assertAllClose(res_actual2, res_desired, rtol=1e-4) 
    self.assertAllClose(res_actual3, res_desired, rtol=1e-4)

  def testMultiplyBatchByScalar(self):
    # Multiply batch of TT-matrix by scalar.
    c = 4.5
    rng = jax.random.PRNGKey(0)
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    right_shape = (4, 4, 4)
    tt = random_.matrix(rng, (left_shape, right_shape), tt_rank=3,
                          batch_shape=(3, 1, 3,), dtype=dtype)
    res_actual1 = ops.full(ops.multiply(tt, c))
    res_actual2 = ops.full(tt * c)
    res_actual3 = ops.full(c * tt)
    res_desired = c * ops.full(tt)
    self.assertAllClose(res_actual1, res_desired, rtol=1e-4) 
    self.assertAllClose(res_actual2, res_desired, rtol=1e-4)  
    self.assertAllClose(res_actual3, res_desired, rtol=1e-4)  


if __name__ == '__main__':
  absltest.main(testLoader=jtu.JaxTestLoader())



"""Utils for compiling functions defined as einsum strings.

Here we use the notion of *tt_einsum*, which is similar to einsum strings, but
with more structure.

Basic element of tt_einsum is a *tt_einsum_core*: list with three elements,
which defines an einsum string for a single TT-core. First element is indices
for the left TT-rank, second element is the indices for the main dimensions of
the resulting TT-core, and the last element is the indices for the right
TT-rank.

TT_einsum consists of list of input and output cores defined like with
tt_einsum_cores.
"""

from typing import Callable, List, Union, Dict
import tree
import opt_einsum as oe
import numpy as np
import jax.numpy as jnp
from string import ascii_lowercase
import copy

from ttax import ops
from ttax.base_class import TT
from ttax.base_class import TTMatrix

# You can use this in TT-einsum expressions. It will be 'i' when working with
# TT-tensors and 'ij' when working with TT-matrices.
I_OR_IJ = 'I_OR_IJ'


class WrappedTT:
  """A class which wraps `TT-object`, which is needed for fusion to work.

  Base `TT-object` class can only have jnp.array objects so that you can pass it into
  jitted function. But, for fusing two functions together we need to track which
  operation created a `TT-object`, so while fusing ops we wrap `TT-objects` with
  this class, to track that.
  """

  def __init__(self, tt: TT, tt_inputs=None, tt_einsum=None):
    self.tt = tt
    self.tt_inputs = tt_inputs
    self.tt_einsum = tt_einsum

  def __mul__(self, other):
    return ops.multiply(self, other)

  def __matmul__(self, other):
    return ops.matmul(self, other)

  def __add__(self, other):
    return ops.add(self, other)

  def __rmul__(self, other):
    return ops.multiply(self, other)

  @property
  def tt_cores(self):
    """Get the list of `TT-cores` of underlying `TT-object`.
    
    :return: `TT-cores`
    :rtype: list
    """
    return self.tt.tt_cores

  @property
  def batch_shape(self):
    """Get the list representing the shape of the batch of underlying `TT-object`. 
    
    :return: batch shape
    :rtype: list
    """
    return self.tt.batch_shape

  @property
  def shape(self):
    """Get the tuple representing the shape of underlying `TT-object`. 
    In batch case includes the shape of the batch.
    
    :return: `TT-Tensor` shape with batch shape
    :rtype: tuple
    """
    return self.tt.shape

  @property
  def axis_dim(self):
    """Get the position of mode axis in underlying `TT-core`.
    It could differ according to the batch shape.
    
    :return: index
    :rtype: int
    """
    return self.tt.axis_dim

  @property
  def num_batch_dims(self):
    """Get the number of batch dimensions for batch of underlying `TT-object`.
    
    :return: number of batch dimensions
    :rtype: int
    """
    return self.tt.num_batch_dims

  @property
  def is_tt_matrix(self):
    """Determine whether the underlying `TT-object` is a `TT-Matrix`.

    :return: `True` if `TT-Matrix`, `False` if `TT-Tensor`
    :rtype: bool
    """
    return self.tt.is_tt_matrix
  
  @property
  def tt_ranks(self):
    """Get `TT-ranks` of the underlying `TT-object` in amount of ``ndim + 1``.
    The first `TT-rank` and the last one equals to `1`.
    
    :return: `TT-ranks`
    :rtype: list
    """
    return self.tt.tt_ranks

  @property
  def ndim(self):
    """Get the number of dimensions of the underlying `TT-object`.
    
    :return: dimensions number
    :rtype: int
    """
    return self.tt.ndim

  @property
  def dtype(self):
    """Represents the `dtype` of elements in underlying `TT-object`.
    
    :return: `dtype` of elements
    :rtype: dtype
    """
    return self.tt.dtype

  @property
  def raw_tensor_shape(self):
    """Get the tuple representing the shape of underlying `TT-object`. 
    Depends on ``raw_tensor_shape``.
    
    :return: shape
    :rtype: list
    """
    return self.tt.raw_tensor_shape

  @property
  def batch_loc(self):
    """Represents the batch indexing for underlying `TT-object`.
    Wraps `TT-object` by special `BatchIndexing` class
    with overloaded ``__getitem__`` method.
    
    Example:
      ``tt.batch_loc[1, :, :]``
    """
    return self.tt.batch_loc


class TTEinsum:
  """A class which contains einsum rule, which is needed for fusion to work.
  """
  def __init__(self, inputs, output, how_to_apply, order='left-to-right'):
    self.inputs = inputs
    self.output = output
    self.how_to_apply = how_to_apply
    self.order = order  # TODO: actually support this.

    if how_to_apply not in ['independent', 'cumulative']:
      raise ValueError('Unsupported "how_to_apply" type "%s"' % how_to_apply)

  def to_vanilla_einsum(self):
    """Build regular einsum."""
    inputs = []
    for inp in self.inputs:
      inputs.append(''.join(inp))
    return ','.join(['...' + a for a in inputs]) + '->...' + ''.join(self.output)

  def apply_mapping(self, mapping: Dict[str, str]):
    """Rename letters according to the given mapping.
    
    :return: new `TTEinsum` with renamed latters
    :rtype: `TTEinsum`
    """
    new_inputs = []
    for inp in self.inputs:
      new_inputs.append(apply_single_mapping(inp, mapping))
    new_output = apply_single_mapping(self.output, mapping)
    return TTEinsum(inputs=new_inputs, output=new_output,
                    how_to_apply=self.how_to_apply)

  def change_input(self, input_idx: int, new_inputs: List):
    """Change argument ``input_idx`` into ``new_inputs``.

    E.g.
      tt_einsum = TTEinsum(inputs=[['a', 'i', 'b'], ['c', 'i', 'd']],
      output=['ac', 'i', 'bd'],
      how_to_apply='independent')
      
      tt_einsum.change_input(0, [['e', 'i', 'f'], ['g', 'i', 'h']])
      
      print(tt_einsum.to_vanilla_einsum())  
      
    will return 'eif,gih,cid->acibd'
    """
    prefix = self.inputs[:input_idx]
    postfix = self.inputs[input_idx + 1:]
    new_inputs = prefix + new_inputs + postfix
    return TTEinsum(new_inputs, self.output, self.how_to_apply)

  def to_distinct_letters(self, distinct_from):
    """Rename letters to make them distinct from letters used in ``distinct_from``."""
    distinct_from_einsum = distinct_from.to_vanilla_einsum()
    # TODO: add upper case
    vacant_letters = [l for l in ascii_lowercase if l not in distinct_from_einsum]
    einsum = self.to_vanilla_einsum()
    curr_unique_letters = [l for l in einsum if l in ascii_lowercase]
    mapping = {}
    for i, l in enumerate(curr_unique_letters):
      mapping[l] = vacant_letters[i]
    return self.apply_mapping(mapping)

  def resolve_i_or_ij(self, is_tt_matrix):
    """Return a version of `TTEinsum` with ``I_OR_IJ`` changed to either 'i' or 'ij'.
    """
    def resolve(el):
      if el == I_OR_IJ:
        return 'ij' if is_tt_matrix else 'i'
      return el
    new_inputs = tree.map_structure(resolve, self.inputs)
    new_output = tree.map_structure(resolve, self.output)
    return TTEinsum(new_inputs, new_output, self.how_to_apply)


def apply_single_mapping(strings, mapping):
  """Apply letter mapping to a list of strings."""
  new_strings = []
  for str in strings:
    curr_str = ''
    for l in str:
      curr_str += mapping.get(l, l)
    new_strings.append(curr_str)
  return new_strings


def _fuse_tt_einsums(tt_einsum: TTEinsum,
                     tensor_args: List[Union[TT, WrappedTT]]) -> TTEinsum:
  """Fuse this TTEinsum with tensor args each of which can be generated by an einsum.

  Let us that you want to fuse the following op:
    flat_inner(a * b, c)
    = sum_{i_1, ..., i_d} a[i_1, ..., i_d] b[i_1, ..., i_d] c[i_1, ..., i_d]
  In this case, your tt_einsum will represent flat_inner
    tt_einsum represents 'aib,cid,ac->bd'
  and the first argument is actually a result of elementwise product, i.e.
    tensor_args[0].tt_einsum represents 'aib,cid->acibd'.

  Then this function will output TTEinsum representing a single fused op
    result represents 'lno,mnp,cnd,lmc->opd'

  It will do so by applying the following 3 steps to every tensor_arg:
  1) change letters of the tensor arg so that they don't intersect with the
      letters of the current einsum. E.g. in this case we will
      change 'aib,cid->acibd' to 'lno,mnp->lmnop': this is the same einsum,
      but it uses letters that are not intersecting with letters in
      the curr_einsum = 'aib,cid,ac->bd'.
  2) change current input of curr_einsum ('aib') into the einsum inputs that
      create this argument ('lno,mnp'). This changes curr_einsum
      from 'aib,cid,ac->bd' into 'lno,mnp,cid,ac->bd'.
  3) make letters in the curr_einsum consistent, e.g. change it
      from 'lno,mnp,cid,ac->bd' to 'lno,mnp,cnd,lmc->opd'. To do that notice
      that in the previous step we changed 'aib' into 'lno,mnp->lmnop', so
      'a' changes to 'lm', 'i' changes to 'n' and 'b' changes to 'op'.
  """
  curr_tt_einsum = copy.deepcopy(tt_einsum)
  curr_tt_einsum_inp_idx = 0
  new_tensor_args = []
  for arg_idx, arg in enumerate(tensor_args):
    if isinstance(arg, WrappedTT) and arg.tt_einsum is not None:
      assert arg.tt_einsum.how_to_apply == 'independent'
      # Step 1.
      new_arg_tt_einsum = arg.tt_einsum.to_distinct_letters(curr_tt_einsum)
      # Step 2.
      unchanged_inp = copy.deepcopy(curr_tt_einsum.inputs[curr_tt_einsum_inp_idx])
      curr_tt_einsum = curr_tt_einsum.change_input(curr_tt_einsum_inp_idx,
                                                   new_arg_tt_einsum.inputs)
      # Step 3.
      mapping = {}
      for fr, to in zip(unchanged_inp, new_arg_tt_einsum.output):
        if len(fr) == 1:
          mapping[fr] = to
        elif len(fr) == len(to):
          for i in range(len(fr)):
            mapping[fr[i]] = to[i]
        else:
          raise ValueError()

      curr_tt_einsum = curr_tt_einsum.apply_mapping(mapping)

      curr_tt_einsum_inp_idx += len(new_arg_tt_einsum.inputs)
      new_tensor_args += arg.tt_inputs
    else:
      curr_tt_einsum_inp_idx += 1
      new_tensor_args.append(arg)

  return curr_tt_einsum, new_tensor_args


def to_function(tt_einsum: TTEinsum) -> Callable:
  """Compile TT-einsum into a function.

  Example:
  
  def multiply(a, b):
  
    tt_einsum = TTEinsum(
    inputs=[['a', 'i', 'b'], ['c', 'i', 'd']],
    output=['ac', 'i', 'bd'],
    how_to_apply='independent'
    )
    
    func = tt_einsum.to_function()
    
    return func(a, b)
  """
  if tt_einsum.how_to_apply == 'independent':
    return compile_independent(tt_einsum)
  elif tt_einsum.how_to_apply == 'cumulative':
    return compile_cumulative(tt_einsum)


def compile_independent(tt_einsum: TTEinsum) -> Callable:
  def new_func(*args):
    are_tt_matrix_inputs = args[0].is_tt_matrix
    tt_einsum_ = tt_einsum.resolve_i_or_ij(are_tt_matrix_inputs)

    is_fusing = any([isinstance(tt, WrappedTT) for tt in args])
    if is_fusing:
      # Have to use a different name to make upper level tt_einsum visible.
      tt_einsum_, args = _fuse_tt_einsums(tt_einsum_, args)
    einsum = tt_einsum_.to_vanilla_einsum()
    num_batch_dims = args[0].num_batch_dims
    # TODO: support broadcasting.
    res_batch_shape = list(args[0].batch_shape)
    # TODO: do in parallel w.r.t. cores.
    # TODO: use optimal einsum.
    res_cores = []
    for i in range(len(args[0].tt_cores)):
      curr_input_cores = [tt.tt_cores[i] for tt in args]
      core = oe.contract(einsum, *curr_input_cores, backend='jax')
      shape = core.shape[num_batch_dims:]
      num_left_rank_dims = len(tt_einsum_.output[0])
      num_tensor_dims = len(tt_einsum_.output[1])
      split_points = (num_left_rank_dims, num_left_rank_dims + num_tensor_dims)
      new_shape = np.split(shape, split_points)
      left_rank = np.prod(new_shape[0])
      right_rank = np.prod(new_shape[2])
      new_shape = [left_rank] + new_shape[1].tolist() + [right_rank]
      new_shape = res_batch_shape + new_shape
      res_cores.append(core.reshape(new_shape))

    if are_tt_matrix_inputs:
      res = TTMatrix(res_cores)
    else:
      res = TT(res_cores)

    if is_fusing:
      res = WrappedTT(res, args, tt_einsum_)
    return res

  return new_func


def compile_cumulative(tt_einsum: TTEinsum) -> Callable:
  def new_func(*args):
    are_tt_matrix_inputs = args[0].is_tt_matrix
    tt_einsum_ = tt_einsum.resolve_i_or_ij(are_tt_matrix_inputs)

    is_fusing = any([isinstance(tt, WrappedTT) for tt in args])
    if is_fusing:
      # Have to use a different name to make upper level tt_einsum visible.
      tt_einsum_, args = _fuse_tt_einsums(tt_einsum_, args)
    einsum = tt_einsum_.to_vanilla_einsum()

    res = jnp.ones([1] * len(args), args[0].tt_cores[0].dtype)
    res_list = []
    for core_idx in range(len(args[0].tt_cores)):
      curr_tensors = [a.tt_cores[core_idx] for a in args]
      curr_tensors.append(res)
      res = oe.contract(einsum, *curr_tensors, backend='jax',
                        optimize='optimal')
      res_list.append(res)
    return res_list

  return new_func


def fuse(func):
  """Fuse a composite function to make it faster.

  Example:
  
  Let's look at f(a, b, c) = <a * b, c> = sum_{i_1, ..., i_d} a[i_1, ..., i_d] b[i_1, ..., i_d] c[i_1, ..., i_d], as ttax.flat_inner(a * b, c) do.

  Function `f` can be suboptimal for some inputs. For example, if `a` and `b`
  are of large TT-rank, and `c` is of low TT-rank, implementing the same
  operation as
  
    ttax.flat_inner(a * c, b)
    
  would be much more efficient.

  `fuse` automates such optimizations. You can build an optimal implementation
  of this function for any inputs by doing
  
    faster_f = ttax.fuse(f)
    
  Finally, don't forget that in JAX to get good speed you need to wrap you
  highest level function in jit, e.g.
  
    faster_f = jax.jit(faster_f)
    
  Now, by using `faster_f(a, b, c)` instead of `f(a, b, c)` you can achieve
  a much faster cumulative time for any inputs.
  """

  def _func(*args):
    wrapped_args = [WrappedTT(arg) for arg in args]
    res = func(*wrapped_args)
    if isinstance(res, WrappedTT):
      res = res.tt
    return res

  return _func


def unwrap_tt(arg):
  """Unwraps argument if it is of `WrappedTT` class, 
  otherwise just returns the argument.

  :type arg: `WrappedTT` or `TTTensOrMat`
  :param arg: argument to unwrap
  :rtype: `TTTensOrMat`
  :return: unwrapped argument
  """
  if isinstance(arg, WrappedTT):
    return arg.tt
  return arg



from absl.testing import absltest
from absl.testing import parameterized

import numpy as np
import jax
import jax.numpy as jnp
import jax.test_util as jtu
from jax.config import config

from ttax.base_class import TT
from ttax.base_class import TTMatrix
from ttax import random_
from ttax import ops
from ttax import autodiff

config.parse_flags_with_absl()


class AutodiffTest(jtu.JaxTestCase):

  def testNormGrad(self):
    dtype = jnp.float32
    rng = jax.random.PRNGKey(42)
    tensor = random_.tensor(rng, (2, 1, 3, 4), tt_rank=[1, 2, 4, 3, 1],
                            dtype=dtype)
    def f(x):
      return 0.5 * ops.flat_inner(x, x)

    grad = autodiff.grad(f)
    actual = grad(tensor)
    desired = tensor
    self.assertAllClose(ops.full(actual), ops.full(desired), rtol=1e-4)

  def testDoubleProjection(self):
    """Compare P grad f(x) against P grad (<x, stop_grad(P grad f(x))>)."""
    dtype = jnp.float32
    rng = jax.random.PRNGKey(42)
    vector = random_.matrix(rng, ((2, 1, 3, 4), (1, 1, 1, 1)),
                            tt_rank=[1, 2, 4, 3, 1], dtype=dtype)
    matrix = random_.matrix(rng, ((2, 1, 3, 4), (2, 1, 3, 4)),
                            tt_rank=[1, 2, 4, 3, 1], dtype=dtype)

    project = autodiff.project(matrix @ vector, vector)
    double_project = autodiff.project(project, vector)
    self.assertAllClose(ops.full(project), ops.full(double_project), rtol=1e-4)

if __name__ == '__main__':
  absltest.main(testLoader=jtu.JaxTestLoader())



from typing import List, Union
import numpy as np
import jax.numpy as jnp
import flax


class TTBase:
  """Represents the base for both `TT-Tensor` and `TT-Matrix` (`TT-object`).
  Includes some basic routines and properties.
  """
  def __mul__(self, other):
    # We can't import ops in the beginning since it creates cyclic dependencies.
    from ttax import ops
    return ops.multiply(self, other)

  def __matmul__(self, other):
    # We can't import ops in the beginning since it creates cyclic dependencies.
    from ttax import ops
    return ops.matmul(self, other)
  
  def __add__(self, other):
    # We can't import ops in the beginning since it creates cyclic dependencies.
    from ttax import ops
    return ops.add(self, other)

  def __rmul__(self, other):
    # We can't import ops in the beginning since it creates cyclic dependencies.
    from ttax import ops
    return ops.multiply(self, other)

  @property
  def axis_dim(self):
    """Get the position of mode axis in `TT-core`.
    It could differ according to the batch shape.
    
    :return: index
    :rtype: int
    """
    return self.num_batch_dims + 1

  @property
  def batch_shape(self):
    """Get the list representing the shape of the batch of `TT-object`. 
    
    :return: batch shape
    :rtype: list
    """
    return self.tt_cores[0].shape[:self.num_batch_dims]

  @property
  def tt_ranks(self):
    """Get `TT-ranks` of the `TT-object` in amount of ``ndim + 1``.
    The first `TT-rank` and the last one equals to `1`.
    
    :return: `TT-ranks`
    :rtype: list
    """
    ranks = [c.shape[self.num_batch_dims] for c in self.tt_cores]
    ranks.append(self.tt_cores[-1].shape[-1])
    return ranks
  
  @property
  def ndim(self):
    """Get the number of dimensions of the `TT-object`.
    
    :return: dimensions number
    :rtype: int
    """
    return len(self.tt_cores)

  @property
  def dtype(self):
    """Represents the `dtype` of elements in `TT-object`.
    
    :return: `dtype` of elements
    :rtype: dtype
    """
    return self.tt_cores[0].dtype

  @property
  def batch_loc(self):
    """Represents the batch indexing for `TT-object`.
    Wraps `TT-object` by special `BatchIndexing` class
    with overloaded ``__getitem__`` method.
    
    Example:
      ``tt.batch_loc[1, :, :]``
    """
    return BatchIndexing(self)



@flax.struct.dataclass
class TT(TTBase):
  """Represents a `TT-Tensor` object as a list of `TT-cores`.
  
  `TT-Tensor` cores take form (r_l, n, r_r), where
  
  - r_l, r_r are `TT-ranks`
  - n makes `TT-Tensor` shape
  """
  tt_cores: List[jnp.array]

  @property
  def shape(self):
    """Get the tuple representing the shape of `TT-Tensor`. 
    In batch case includes the shape of the batch.
    
    :return: `TT-Tensor` shape with batch shape
    :rtype: tuple
    """
    no_batch_shape = [c.shape[self.axis_dim] for c in self.tt_cores]
    return tuple(list(self.batch_shape) + no_batch_shape)

  @property
  def num_batch_dims(self):
    """Get the number of batch dimensions for batch of `TT-Tensors`.
    
    :return: number of batch dimensions
    :rtype: int
    """
    return len(self.tt_cores[0].shape) - 3

  @property
  def is_tt_matrix(self):
    """Determine whether the object is a `TT-Matrix`.

    :return: `True` if `TT-Matrix`, `False` if `TT-Tensor`
    :rtype: bool
    """
    return False

  @property
  def raw_tensor_shape(self):
    """Get the tuple representing the shape of `TT-Tensor`. 
    In batch case does not include the shape of the batch.
    
    :return: `TT-Tensor` shape
    :rtype: tuple
    """
    return [c.shape[self.axis_dim] for c in self.tt_cores]
  
  def __str__(self):
    """Creates a string describing TT-Tensor.
    :return: TT-Tensor description
    :rtype: string
    """
    if self.num_batch_dims == 0:
      s = "TT-Tensor of shape {0} and TT-ranks {1}"
      s = s.format(self.shape, self.tt_ranks)
    else:
      s = "Batch of {0} TT-Tensors of shape {1} and TT-ranks {2}"
      s = s.format(self.batch_shape, self.raw_tensor_shape, self.tt_ranks)
    s += " with {0} elements.".format(self.dtype)
    return s
  
  def __getitem__(self, slice_spec):
    """Basic indexing, returns a TT containing the specified element / slice.
    
    Examples:
      >>> a = ttax.random.tensor(rng, [2, 3, 4])
      >>> a[1, :, :]
      is a 2D TensorTrain 3 x 4.
      >>> a[1:2, :, :]
      is a 3D TensorTrain 1 x 3 x 4
    """
    if len(slice_spec) != self.ndim:
      raise ValueError('Expected %d indices, got %d' % (self.ndim,
                                                        len(slice_spec)))
    new_tt_cores = []
    remainder = None
    for i in range(self.ndim):
      curr_core = self.tt_cores[i]
      sliced_core = curr_core[..., :, slice_spec[i], :]
      if len(curr_core.shape) != len(sliced_core.shape):
        # This index is specified exactly and we want to collapse this axis.
        if remainder is None:
          remainder = sliced_core
        else:
          remainder = jnp.matmul(remainder, sliced_core)
      else:
        if remainder is not None:
          # Add reminder from the previous collapsed cores to the current core.
          sliced_core = jnp.einsum('...ab,...bid->...aid', 
                                   remainder, sliced_core)
          remainder = None
        new_tt_cores.append(sliced_core)

    if remainder is not None:
      # The reminder obtained from collapsing the last cores.
      new_tt_cores[-1] = jnp.einsum('...aib,...bd->...aid', 
                                    new_tt_cores[-1], remainder)
      remainder = None
    return TT(new_tt_cores)


@flax.struct.dataclass
class TTMatrix(TTBase):
  """Represents a `TT-Matrix` object as a list of `TT-cores`.
  
  `TT-Matrix` cores take form (r_l, n_l, n_r, r_r), where
  
  - r_l, r_r are `TT-ranks` just as for `TT-Tensor`
  - n_l, n_r make left and right shapes of `TT-Matrix` as rows and cols
  """
  tt_cores: List[jnp.array]

  @property
  def raw_tensor_shape(self):
    """Get the lists representing left and right shapes of `TT-Matrix`. 
    In batch case does not include the shape of the batch.
    
    For example if `TT-Matrix` cores are (1, 2, 3, 5) (5, 6, 7, 1)
    returns (2, 6), (3, 7).
    
    :return: `TT-Matrix` shapes
    :rtype: list, list
    """
    left_shape = [c.shape[self.axis_dim] for c in self.tt_cores]
    right_shape = [c.shape[self.axis_dim + 1] for c in self.tt_cores]
    return left_shape, right_shape

  @property
  def shape(self):
    """Get the tuple representing the shape of underlying dense tensor as matrix. 
    In batch case includes the shape of the batch.
    
    For example if `TT-Matrix` cores are (1, 2, 3, 5) (5, 6, 7, 1)
    it's shape is (12, 21).
    
    :return: `TT-Matrix` shape in dense form with batch shape
    :rtype: tuple
    """
    left_shape, right_shape = self.raw_tensor_shape
    no_batch_shape = [np.prod(left_shape), np.prod(right_shape)]
    return tuple(list(self.batch_shape) + no_batch_shape)

  @property
  def num_batch_dims(self):
    """Get the number of batch dimensions for batch of `TT-Matrices.`
    
    :return: number of batch dimensions
    :rtype: int
    """
    return len(self.tt_cores[0].shape) - 4

  @property
  def is_tt_matrix(self):
    """Determine whether the object is a `TT-Matrix`.

    :return: `True` if `TT-Matrix`, `False` if `TT-Tensor`
    :rtype: bool
    """
    return True
  
  def __str__(self):
    """Creates a string describing TT-Matrix.
    :return: TT-Matrix description
    :rtype: string
    """
    if self.num_batch_dims == 0:
      s = "TT-Matrix of shape {0} and TT-ranks {1}"
      s = s.format(self.raw_tensor_shape, self.tt_ranks)
    else:
      s = "Batch of {0} TT-Matrices of shape {1} and TT-ranks {2}"
      s = s.format(self.batch_shape, self.raw_tensor_shape, self.tt_ranks)
    s += " with {0} elements.".format(self.dtype)
    return s
  
  def __getitem__(self, slice_spec):
    """Basic indexing, returns a TTMatrix containing the specified element / slice."""
    d = self.ndim
    if len(slice_spec) != 2 * d:
      raise ValueError('Expected %d indices, got %d' % (2 * d, len(slice_spec)))
    for i in range(d):
      if isinstance(slice_spec[i], slice) != isinstance(slice_spec[d+i], slice):
        raise ValueError('Elements i_%d and j_%d should be the same type, '
                         'instead: %s and %s.' % (i, i, slice_spec[i], 
                                                  slice_spec[d+i]))
    new_tt_cores = []
    remainder = None
    for i in range(self.ndim):
      curr_core = self.tt_cores[i]
      sliced_core = curr_core[..., :, slice_spec[i], slice_spec[d+i], :]
      if len(curr_core.shape) != len(sliced_core.shape):
        # These indices are specified exactly and we want to collapse this axis.
        if remainder is None:
          remainder = sliced_core
        else:
          remainder = jnp.matmul(remainder, sliced_core)
      else:
        if remainder is not None:
          # Add reminder from the previous collapsed cores to the current core.
          sliced_core = jnp.einsum('...ab,...bijd->...aijd', 
                                   remainder, sliced_core)
          remainder = None
        new_tt_cores.append(sliced_core)

    if remainder is not None:
      # The reminder obtained from collapsing the last cores.
      new_tt_cores[-1] = jnp.einsum('...aijb,...bd->...aijd', 
                                    new_tt_cores[-1], remainder)
      remainder = None

    return TTMatrix(new_tt_cores)


class BatchIndexing:
  def __init__(self, tt):
    self._tt = tt

  def __getitem__(self, indices: list):
    non_none_indices = [idx for idx in indices if idx is not None]
    if len(non_none_indices) > self._tt.num_batch_dims:
      raise ValueError('Expected %d indices, got %d' % (self._tt.num_batch_dims,
                                                        len(non_none_indices)))
    new_cores = []
    for core_idx in range(self._tt.ndim):
      curr_core = self._tt.tt_cores[core_idx]
      new_cores.append(curr_core.__getitem__(indices))

    if self._tt.is_tt_matrix:
      return TTMatrix(new_cores)
    else:
      return TT(new_cores)

TTTensOrMat = Union[TT, TTMatrix]



import jax
import jax.numpy as jnp

from ttax.base_class import TTBase
from ttax.base_class import TT
from ttax.base_class import TTMatrix
from ttax.compile import WrappedTT

def is_tt_tensor(arg) -> bool:
  """Determine whether the object is a `TT-Tensor` or `WrappedTT` with underlying `TT-Tensor`.
  
  :return: `True` if `TT-Tensor` or `WrappedTT(TT-Tensor)`, `False` otherwise
  :rtype: bool
  """
  return isinstance(arg, TT) or (isinstance(arg, WrappedTT) and 
         not arg.tt.is_tt_matrix)

def is_tt_matrix(arg) -> bool:
  """Determine whether the object is a `TT-Matrix` or `WrappedTT` with underlying `TT-Matrix`.
  
  :return: `True` if `TT-Matrix` or `WrappedTT(TT-Matrix)`, `False` otherwise
  :rtype: bool
  """
  return isinstance(arg, TTMatrix) or (isinstance(arg, WrappedTT) and
         arg.is_tt_matrix)

def is_tt_object(arg) -> bool:
  """Determine whether the object is a `TT-Tensor`, `TT-Matrix` or `WrappedTT` with one of them.
  
  :return: `True` if `TT-object`, `False` otherwise
  :rtype: bool
  """
  return is_tt_tensor(arg) or is_tt_matrix(arg)



from ttax.base_class import TTBase
from ttax.base_class import TT
from ttax.base_class import TTMatrix

from ttax.ops import full
from ttax.ops import multiply
from ttax.ops import flat_inner
from ttax.ops import matmul
from ttax.ops import add
from ttax.ops import tt_vmap
from ttax.ops import are_shapes_equal
from ttax.ops import are_batches_broadcastable

from ttax.compile import fuse
from ttax.compile import unwrap_tt
from ttax.compile import to_function
from ttax.compile import I_OR_IJ
from ttax.compile import WrappedTT
from ttax.compile import TTEinsum

from ttax.decompositions import round
from ttax.decompositions import orthogonalize

from ttax.riemannian import tangent_to_deltas

from ttax import random_ as random
from ttax.random_ import tensor
from ttax.random_ import matrix

from ttax.utils import is_tt_tensor
from ttax.utils import is_tt_matrix
from ttax.utils import is_tt_object

from ttax.autodiff import grad
from ttax.autodiff import hessian_vector_product
from ttax.autodiff import project



import numpy as np
import jax
import jax.numpy as jnp

from ttax.base_class import TT
from ttax.base_class import TTMatrix


def tensor(rng, shape, tt_rank=2, batch_shape=None, dtype=jnp.float32):
  """Generate a random `TT-Tensor` of the given shape and `TT-rank`.
  
  :param rng: JAX PRNG key
  :type rng: random state is described by two unsigned 32-bit integers
  :param shape: desired tensor shape
  :type shape: array
  :param tt_rank: desired `TT-ranks` of `TT-Tensor`
  :type tt_rank: single number for equal `TT-ranks` or array specifying all `TT-ranks`
  :param batch_shape: desired batch shape of `TT-Tensor`
  :type batch_shape: array
  :param dtype: type of elements in `TT-Tensor`
  :type dtype: `dtype`
  :return: generated `TT-Tensor`
  :rtype: TT
  """
  shape = np.array(shape)
  tt_rank = np.array(tt_rank)
  batch_shape = list(batch_shape) if batch_shape else []

  num_dims = shape.size
  if tt_rank.size == 1:
    tt_rank = tt_rank * np.ones(num_dims - 1)
    tt_rank = np.insert(tt_rank, 0, 1)
    tt_rank = np.append(tt_rank, 1)

  tt_rank = tt_rank.astype(int)

  tt_cores = []
  rng_arr = jax.random.split(rng, num_dims)
  for i in range(num_dims):
    curr_core_shape = [tt_rank[i], shape[i], tt_rank[i + 1]]
    curr_core_shape = batch_shape + curr_core_shape
    tt_cores.append(jax.random.normal(rng_arr[i], curr_core_shape, dtype=dtype))

  return TT(tt_cores)


def matrix(rng, shape, tt_rank=2, batch_shape=None, dtype=jnp.float32):
  """Generate a random `TT-Matrix` of the given shape and `TT-rank`.
    
  :param rng: JAX PRNG key
  :type rng: random state is described by two unsigned 32-bit integers
  :param shape: desired tensor shape
  :type shape: array
  :param tt_rank: desired `TT-ranks` of `TT-Matrix`
  :type tt_rank: single number for equal `TT-ranks` or array specifying all `TT-ranks`
  :param batch_shape: desired batch shape of `TT-Matrix`
  :type batch_shape: array
  :param dtype: type of elements in `TT-Matrix`
  :type dtype: `dtype`
  :return: generated `TT-Matrix`
  :rtype: TTMatrix
  """
  shape = [np.array(shape[0]), np.array(shape[1])]
  tt_rank = np.array(tt_rank)
  batch_shape = list(batch_shape) if batch_shape else []

  num_dims = shape[0].size
  if tt_rank.size == 1:
    tt_rank = tt_rank * np.ones(num_dims - 1)
    tt_rank = np.insert(tt_rank, 0, 1)
    tt_rank = np.append(tt_rank, 1)

  tt_rank = tt_rank.astype(int)

  tt_cores = []
  rng_arr = jax.random.split(rng, num_dims)
  for i in range(num_dims):
    curr_core_shape = [tt_rank[i], shape[0][i], shape[1][i], tt_rank[i + 1]]
    curr_core_shape = batch_shape + curr_core_shape
    tt_cores.append(jax.random.normal(rng_arr[i], curr_core_shape, dtype=dtype))

  return TTMatrix(tt_cores)



from absl.testing import absltest
from absl.testing import parameterized

import time
import numpy as np
import jax
import jax.numpy as jnp
import jax.test_util as jtu
from jax.config import config

from ttax import random_
from ttax import ops
from ttax import compile

config.parse_flags_with_absl()


def benchmark(func, *args):
  func_jit = jax.jit(func)
  # Warmup.
  for _ in range(3):
    func_jit(*args).block_until_ready()

  num_repeats = 100
  s = time.time()
  outs = []
  for _ in range(num_repeats):
    outs.append(func_jit(*args))
  for o in outs:
    o.block_until_ready()
  e = time.time()
  return (e - s) / num_repeats


class TTTensorTest(jtu.JaxTestCase):

  @parameterized.named_parameters(
      ('_a*b*c', 'a*b*c'),
      ('<a*b, c>', '<a*b, c>')
  )
  def testFuse(self, op_type):
    np.random.seed(1)
    rng1, rng2, rng3 = jax.random.split(jax.random.PRNGKey(0), 3)
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (1, 2, 3, 4), tt_rank=2, dtype=dtype)
    tt_b = random_.tensor(rng2, (1, 2, 3, 4), tt_rank=[1, 1, 4, 3, 1], dtype=dtype)
    tt_c = random_.tensor(rng3, (1, 2, 3, 4), tt_rank=3, dtype=dtype)
    if op_type == 'a*b*c':
      func = lambda a, b, c: ops.multiply(ops.multiply(a, b), c)
      fused_func = compile.fuse(func)
      res_actual = ops.full(func(tt_a, tt_b, tt_c))
      res_desired = ops.full(fused_func(tt_a, tt_b, tt_c))
    elif op_type == '<a*b, c>':
      func = lambda a, b, c: ops.flat_inner(ops.multiply(a, b), c)
      fused_func = compile.fuse(func)
      res_actual = func(tt_a, tt_b, tt_c)
      res_desired = fused_func(tt_a, tt_b, tt_c)

      self.assertAllClose(res_actual, res_desired)

  def testFuseIsFaster(self):
    np.random.seed(1)
    rng1, rng2, rng3 = jax.random.split(jax.random.PRNGKey(0), 3)
    dtype = jnp.float32
    tt_a = random_.tensor(rng1, (10, 10, 10, 10), tt_rank=30, dtype=dtype)
    tt_b = random_.tensor(rng2, (10, 10, 10, 10), tt_rank=30, dtype=dtype)
    tt_c = random_.tensor(rng3, (10, 10, 10, 10), tt_rank=1, dtype=dtype)
    func = lambda a, b, c: ops.flat_inner(ops.multiply(a, b), c)
    fused_func = compile.fuse(func)

    func_speed = benchmark(func, tt_a, tt_b, tt_c)
    fused_func_speed = benchmark(fused_func, tt_a, tt_b, tt_c)
    # Check that fused version is at least 10x faster than non-fused.
    self.assertLess(fused_func_speed, 0.1 * func_speed)


class TTMatrixTest(jtu.JaxTestCase):

  @parameterized.named_parameters(
      ('(a*b) @ c', '(a*b) @ c'),
  )
  def testFuse(self, op_type):
    np.random.seed(1)
    rng1, rng2, rng3 = jax.random.split(jax.random.PRNGKey(0), 3)
    dtype = jnp.float32
    left_shape = (2, 3, 4)
    sum_shape = (4, 3, 5)
    right_shape = (4, 4, 4)
    tt_a = random_.matrix(rng1, (left_shape, sum_shape), tt_rank=3, dtype=dtype)
    tt_b = random_.matrix(rng2, (left_shape, sum_shape), tt_rank=3, dtype=dtype)
    tt_c = random_.matrix(rng3, (sum_shape, right_shape), tt_rank=[1, 4, 3, 1],
                              dtype=dtype)
    if op_type == '(a*b) @ c':
      def func(a, b, c):
        return (a * b) @ c
      fused_func = compile.fuse(func)
      res_actual = ops.full(func(tt_a, tt_b, tt_c))
      res_desired = ops.full(fused_func(tt_a, tt_b, tt_c))
      self.assertAllClose(res_actual, res_desired, rtol=1e-4)


if __name__ == '__main__':
  absltest.main(testLoader=jtu.JaxTestLoader())


import numpy as np
import jax
import jax.numpy as jnp

from ttax.base_class import TT
from ttax.base_class import TTMatrix
from ttax.ops import tt_vmap


@tt_vmap(1)
def round(tt, max_tt_rank=None, epsilon=None):
  """Tensor Train rounding procedure, returns a `TT-object` with smaller `TT-ranks`.

  :param tt: argument which ranks would be reduced
  :type tt: `TT-Tensor` or `TT-Matrix`
  :type max_tt_rank: int or list of ints
  :param max_tt_rank: 
  
    - If a number, than defines the maximal `TT-rank` of the result. 
    - If a list of numbers, than `max_tt_rank` length should be d+1 
      (where d is the number of dimensions) and `max_tt_rank[i]`
      defines the maximal (i+1)-th `TT-rank` of the result.
      
      The following two versions are equivalent
      
      - ``max_tt_rank = r``
      
      - ``max_tt_rank = [1] + [r] * (d-1) + [1]``  
       
  :type epsilon: float or None
  :param epsilon:
  
    - If the `TT-ranks` are not restricted (`max_tt_rank=None`), then
      the result would be guarantied to be `epsilon`-close to `tt`
      in terms of relative Frobenius error:
      
        `||res - tt||_F / ||tt||_F <= epsilon`
        
    - If the `TT-ranks` are restricted, providing a loose `epsilon` may
      reduce the `TT-ranks` of the result. E.g.
      
        ``round(tt, max_tt_rank=100, epsilon=0.9)``
        
      will probably return you a `TT-Tensor` with `TT-ranks` close to 1, not 100.
      Note that providing a nontrivial (= not equal to `None`) epsilon will make
      the `TT-ranks` of the result change depending on the data, 
      which will prevent you from using ``jax.jit``
      for speeding up the computations.
  :return: `TT-object` with reduced `TT-ranks`
  :rtype: `TT-Tensor` or `TT-Matrix`
  :raises:
    ValueError if `max_tt_rank` is less than 0, if `max_tt_rank` is not a number and
    not a vector of length d + 1 where d is the number of dimensions of
    the input tensor, if `epsilon` is less than 0.
  """

  if max_tt_rank is None:
    max_tt_rank = np.iinfo(np.int32).max
  num_dims = tt.ndim
  max_tt_rank = np.array(max_tt_rank).astype(np.int32)
  if np.any(max_tt_rank < 1):
    raise ValueError('Maximum TT-rank should be greater or equal to 1.')
  if epsilon is not None:
    raise NotImplementedError('Epsilon is not supported yet.')
  if max_tt_rank.size == 1:
    max_tt_rank = (max_tt_rank * np.ones(num_dims + 1)).astype(jnp.int32)
  elif max_tt_rank.size != num_dims + 1:
    raise ValueError('max_tt_rank should be a number or a vector of size (d+1) '
                     'where d is the number of dimensions of the tensor.')
  if tt.is_tt_matrix:
    raw_shape = tt.raw_tensor_shape
  else:
    raw_shape = tt.shape

  tt_cores = orthogonalize(tt).tt_cores
  # Copy cores references so we can change the cores.
  tt_cores = list(tt_cores)

  ranks = [1] * (num_dims + 1)

  # Right to left SVD compression.
  for core_idx in range(num_dims - 1, 0, -1):
    curr_core = tt_cores[core_idx]
    if tt.is_tt_matrix:
      curr_mode_left = raw_shape[0][core_idx]
      curr_mode_right = raw_shape[1][core_idx]
      curr_mode = curr_mode_left * curr_mode_right
    else:
      curr_mode = raw_shape[core_idx]
    
    columns = curr_mode * ranks[core_idx + 1]
    curr_core = jnp.reshape(curr_core, [-1, columns])
    rows = curr_core.shape[0]
    if max_tt_rank[core_idx] == 1:
      ranks[core_idx] = 1
    else:
      ranks[core_idx] = min(max_tt_rank[core_idx], rows, columns)
    u, s, v = jnp.linalg.svd(curr_core, full_matrices=False)
    u = u[:, 0:ranks[core_idx]]
    s = s[0:ranks[core_idx]]
    v = v[0:ranks[core_idx], :]
    if tt.is_tt_matrix:
      core_shape = (ranks[core_idx], curr_mode_left, curr_mode_right,
                    ranks[core_idx + 1])
    else:
      core_shape = (ranks[core_idx], curr_mode, ranks[core_idx + 1])
    tt_cores[core_idx] = jnp.reshape(v, core_shape)
    prev_core_shape = (-1, rows)
    tt_cores[core_idx - 1] = jnp.reshape(tt_cores[core_idx - 1], prev_core_shape)
    tt_cores[core_idx - 1] = jnp.matmul(tt_cores[core_idx - 1], u)
    tt_cores[core_idx - 1] = jnp.matmul(tt_cores[core_idx - 1], jnp.diag(s))

  if tt.is_tt_matrix:
    core_shape = (ranks[0], raw_shape[0][0], raw_shape[1][0], ranks[1])
  else:
    core_shape = (ranks[0], raw_shape[0], ranks[1])
  tt_cores[0] = jnp.reshape(tt_cores[0], core_shape)

  if tt.is_tt_matrix:
    return TTMatrix(tt_cores)
  else:
    return TT(tt_cores)


def orthogonalize(tt, left_to_right=True):
  """Orthogonalize `TT-cores` of a `TT-object`.
  
  :type tt: `TT-Tensor` or `TT-Matrix`
  :param tt: `TT-object` which `TT-cores` would be orthogonalized
  :param left_to_right: the direction of orthogonalization, `True` for left to right and `False` for right to left
  :type left_to_right: bool
  :return: `TT-object` with orthogonalized `TT-cores`
  :rtype: `TT-Tensor` or `TT-Matrix`
  """
  if left_to_right:
    return _orthogonalize_tt_cores_left_to_right(tt)
  else:
    return _orthogonalize_tt_cores_right_to_left(tt)


def _orthogonalize_tt_cores_left_to_right(tt):
  """Orthogonalize TT-cores of a TT-object.
  Args:
    tt: TT-tensor or TT-matrix.
    TT-tensor or TT-matrix.
  """

  # Left to right orthogonalization.
  num_dims = tt.ndim
  if tt.is_tt_matrix:
    raw_shape = tt.raw_tensor_shape
  else:
    raw_shape = tt.shape

  tt_ranks = tt.tt_ranks
  next_rank = tt_ranks[0]
  # Copy cores references so we can change the cores.
  tt_cores = list(tt.tt_cores)
  for core_idx in range(num_dims - 1):
    curr_core = tt_cores[core_idx]
    # TT-ranks could have changed on the previous iteration, so `tt_ranks` can
    # be outdated for the current TT-rank, but should be valid for the next
    # TT-rank.
    curr_rank = next_rank
    next_rank = tt_ranks[core_idx + 1]
    if tt.is_tt_matrix:
      curr_mode_left = raw_shape[0][core_idx]
      curr_mode_right = raw_shape[1][core_idx]
      curr_mode = curr_mode_left * curr_mode_right
    else:
      curr_mode = raw_shape[core_idx]
    
    qr_shape = (curr_rank * curr_mode, next_rank)
    curr_core = jnp.reshape(curr_core, qr_shape)
    curr_core, triang = jnp.linalg.qr(curr_core)
    
    triang_shape = triang.shape

    # The TT-rank could have changed: if qr_shape is e.g. 4 x 10, than q would
    # be of size 4 x 4 and r would be 4 x 10, which means that the next rank
    # should be changed to 4.
    next_rank = triang_shape[0]
    if tt.is_tt_matrix:
      new_core_shape = (curr_rank, curr_mode_left, curr_mode_right, next_rank)
    else:
      new_core_shape = (curr_rank, curr_mode, next_rank)
    tt_cores[core_idx] = jnp.reshape(curr_core, new_core_shape)

    next_core = jnp.reshape(tt_cores[core_idx + 1], (triang_shape[1], -1))
    tt_cores[core_idx + 1] = jnp.matmul(triang, next_core)

  if tt.is_tt_matrix:
    last_core_shape = (next_rank, raw_shape[0][-1], raw_shape[1][-1], 1)
  else:
    last_core_shape = (next_rank, raw_shape[-1], 1)
  tt_cores[-1] = jnp.reshape(tt_cores[-1], last_core_shape)

  if tt.is_tt_matrix:
    return TTMatrix(tt_cores)
  else:
    return TT(tt_cores)


def _orthogonalize_tt_cores_right_to_left(tt):
  """Orthogonalize TT-cores of a TT-object.
  Args:
    tt: TT-tensor or TT-matrix.
  Returns:
    TT-tensor or TT-matrix.
  """
  
  # Right to left orthogonalization.
  num_dims = tt.ndim
  if tt.is_tt_matrix:
    raw_shape = tt.raw_tensor_shape
  else:
    raw_shape = tt.shape

  tt_ranks = tt.tt_ranks
  prev_rank = tt_ranks[num_dims]
  # Copy cores references so we can change the cores.
  tt_cores = list(tt.tt_cores)
  for core_idx in range(num_dims - 1, 0, -1):
    curr_core = tt_cores[core_idx]
    # TT-ranks could have changed on the previous iteration, so `tt_ranks` can
    # be outdated for the current TT-rank, but should be valid for the next
    # TT-rank.
    curr_rank = prev_rank
    prev_rank = tt_ranks[core_idx]
    if tt.is_tt_matrix:
      curr_mode_left = raw_shape[0][core_idx]
      curr_mode_right = raw_shape[1][core_idx]
      curr_mode = curr_mode_left * curr_mode_right
    else:
      curr_mode = raw_shape[core_idx]
    
    qr_shape = (prev_rank, curr_mode * curr_rank)
    curr_core = jnp.reshape(curr_core, qr_shape)
    curr_core, triang = jnp.linalg.qr(curr_core.T)
    curr_core = curr_core.T
    triang = triang.T
    triang_shape = triang.shape

    # The TT-rank could have changed: if qr_shape is e.g. 4 x 10, than q would
    # be of size 4 x 4 and r would be 4 x 10, which means that the next rank
    # should be changed to 4.
    prev_rank = triang_shape[1]
    if tt.is_tt_matrix:
      new_core_shape = (prev_rank, curr_mode_left, curr_mode_right, curr_rank)
    else:
      new_core_shape = (prev_rank, curr_mode, curr_rank)
    tt_cores[core_idx] = jnp.reshape(curr_core, new_core_shape)

    prev_core = jnp.reshape(tt_cores[core_idx - 1], (-1, triang_shape[0]))
    tt_cores[core_idx - 1] = jnp.matmul(prev_core, triang)

  if tt.is_tt_matrix:
    first_core_shape = (1, raw_shape[0][0], raw_shape[1][0], prev_rank)
  else:
    first_core_shape = (1, raw_shape[0], prev_rank)
  tt_cores[0] = jnp.reshape(tt_cores[0], first_core_shape)
  
  if tt.is_tt_matrix:
    return TTMatrix(tt_cores)
  else:
    return TT(tt_cores)



from absl.testing import absltest

import numpy as np
import jax
import jax.numpy as jnp
import jax.test_util as jtu
from jax.config import config

from ttax import random_
from ttax import ops
from ttax import autodiff
from ttax import riemannian

config.parse_flags_with_absl()


class TTTensorTest(jtu.JaxTestCase):

  def testToAndFromDeltas(self):
    rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
    dtype = jnp.float32
    what = random_.tensor(rng1, (4, 5, 6), tt_rank=4, dtype=dtype)
    where = random_.tensor(rng2, (4, 5, 6), tt_rank=3, dtype=dtype)
    projected = autodiff.project(what, where)

    deltas = riemannian.tangent_to_deltas(projected)
    reconstructed_projected = riemannian.deltas_to_tangent(deltas, where)
    # Tangent space element norm can be computed from deltas norm.
    projected_normsq_desired = ops.flat_inner(projected, projected)
    projected_normsq_actual = sum([jnp.sum(c * c) for c in deltas])
    self.assertAllClose(ops.full(projected), ops.full(reconstructed_projected))
    self.assertAllClose(projected_normsq_desired, projected_normsq_actual)
#
#   def testToAndFromDeltasBatch(self):
#     rng1, rng2 = jax.random.split(jax.random.PRNGKey(0))
#     dtype = jnp.float32
#     what = random_.tensor(rng1, (4, 5, 6), tt_rank=4, dtype=dtype,
#                           batch_shape=(3,))
#     where = random_.tensor(rng2, (4, 5, 6), tt_rank=3, dtype=dtype,
#                            batch_shape=(3,))
#     projected = autodiff.project(what, where)
#
#     deltas = riemannian.tangent_to_deltas(projected)
#     reconstructed_projected = riemannian.deltas_to_tangent(deltas, where)
#     self.assertAllClose(ops.full(projected), ops.full(reconstructed_projected))
#     # Tangent space element norm can be computed from deltas norm.
#     projected_normsq_desired = ops.flat_inner(projected, projected)
#     for i in range(3):
#       projected_normsq_actual = sum([jnp.sum(c[i] * c[i]) for c in deltas])
#       self.assertAllClose(projected_normsq_desired[i], projected_normsq_actual)


if __name__ == '__main__':
  absltest.main(testLoader=jtu.JaxTestLoader())



from setuptools import setup

setup(name='ttax',
      version='0.0.3',
      description='Tensor Train decomposition on Jax',
      url='https://github.com/fasghq/ttax',
      author='Alexander Novikov and Dmitry Belousov',
      author_email='sasha.v.novikov@gmail.com',
      license='MIT',
      packages=['ttax'],
      install_requires=[
          'numpy',
          'dm-tree',
          'jax',
          'flax'
      ],
      zip_safe=False)



import functools
import jax
import jax.numpy as jnp
import numpy as np

from ttax.base_class import TTBase
from ttax.base_class import TT
from ttax.base_class import TTMatrix
from ttax.compile import WrappedTT
from ttax.compile import TTEinsum, to_function, I_OR_IJ
from ttax.compile import unwrap_tt
from ttax.utils import is_tt_tensor
from ttax.utils import is_tt_matrix
from ttax.utils import is_tt_object


def tt_vmap(num_batch_args=None):
  """Decorator which makes a function support batch TT-inputs.
  
  :type num_batch_args: int or None
  :param num_batch_args: The amount of arguments that are batches of `TT-objects`.
  
    - If None, than function will be mapped over all arguments.
    - If integer, specifies the count of first arguments to map over, e.g.
      `num_batch_args=n` means that function will be mapped over 
      first `n` arguments.
  :return: Decorator
  
  Comments:
    The function is vmapped `num_batch_dims` times, as it supports 
    multidimensional batches. The number of batch dimension to be 
    mapped over is shown by `num_batch_dims` property and should be 
    the same for all args of the function, by which it will be mapped over. 
    Otherwise such axis should be specified by `num_batch_args`.
  """
  def tt_vmap_fixed_batching_pattern(func):
    @functools.wraps(func)
    def vectorized_func(*args, **kwargs):
      tt_arg = args[0]  # TODO: what if only kwargs are present?
      if tt_arg.num_batch_dims == 0:
        return func(*args, **kwargs)
      else:
        if num_batch_args is not None:
          num_non_batch_args = len(args) + len(kwargs) - num_batch_args
          in_axis = [0] * num_batch_args + [None] * num_non_batch_args
          num_args = num_batch_args
        else:
          num_args = len(args) + len(kwargs)
          in_axis = [0] * num_args
        if num_args > 1 and (isinstance(args[1], TTMatrix) or
                                 isinstance(args[1], TT)):
          if args[0].is_tt_matrix != args[1].is_tt_matrix:
            raise ValueError('Types of the arguments are different.')
          if not are_batches_broadcastable(args[0], args[1]):
            raise ValueError('The batch sizes are different and not 1, '
                             'broadcasting is not available.')
          broadcast_shape = np.maximum(list(args[0].batch_shape), 
                                       list(args[1].batch_shape))
          new_args = list(args)
          if args[0].is_tt_matrix:
            for i, tt in enumerate(args[:2]):
              new_cores = []
              for core in tt.tt_cores:
                core = jnp.broadcast_to(core, list(broadcast_shape) +
                                              list(core.shape[-4:]))
                new_cores.append(core)
              new_args[i] = TTMatrix(new_cores)
          else:
            for i, tt in enumerate(args[:2]):
              new_cores = []
              for core in tt.tt_cores:
                core = jnp.broadcast_to(core, list(broadcast_shape) + 
                                              list(core.shape[-3:]))
                new_cores.append(core)
              new_args[i] = TT(new_cores)
        else:
          new_args = args
        vmapped = func
        for _ in range(tt_arg.num_batch_dims):
          vmapped = jax.vmap(vmapped, in_axis)
        return vmapped(*new_args, **kwargs)
    return vectorized_func
  return tt_vmap_fixed_batching_pattern


def full_tt_tensor(tt: TT) -> jnp.array:
  """Converts TT into a regular tensor.
  """
  num_dims = tt.ndim

  dtype = tt.tt_cores[0].dtype  # TODO: make tt.dtype.
  res = jnp.ones((1, 1), dtype=dtype)
  for i in range(num_dims):
    curr_core = tt.tt_cores[i]
    right_rank = curr_core.shape[-1]
    res = jnp.einsum('pa,aib->pib', res, curr_core)
    res = res.reshape(-1, right_rank)

  return jnp.reshape(res, tt.shape)


def full_tt_matrix(tt: TTMatrix) -> jnp.array:
  """Converts TT-matrix into a regular matrix.
  """
  num_dims = tt.ndim

  dtype = tt.tt_cores[0].dtype  # TODO: make tt.dtype.
  res = jnp.ones((1, 1), dtype=dtype)
  for i in range(num_dims):
    curr_core = tt.tt_cores[i]
    right_rank = curr_core.shape[-1]
    res = jnp.einsum('pa,aijb->pijb', res, curr_core)
    res = res.reshape(-1, right_rank)

  raw_shape = tt.raw_tensor_shape
  intermediate_shape = []
  for i in range(num_dims):
    intermediate_shape.append(raw_shape[0][i])
    intermediate_shape.append(raw_shape[1][i])
  res = jnp.reshape(res, list(tt.batch_shape) + intermediate_shape)
  transpose = []
  for i in range(0, 2 * num_dims, 2):
    transpose.append(i)
  for i in range(1, 2 * num_dims, 2):
    transpose.append(i)
  res = jnp.transpose(res, transpose)
  return jnp.reshape(res, tt.shape)


@tt_vmap()
def full(tt: TTBase) -> jnp.array:
  """Converts `TT-Tensor` or `TT-Matrix` into a dense format.
  """
  if isinstance(tt, TT):
    return full_tt_tensor(tt)
  elif isinstance(tt, TTMatrix):
    return full_tt_matrix(tt)


def tt_tt_multiply(a, b):
  tt_einsum = TTEinsum(
      inputs=[['a', I_OR_IJ, 'b'], ['c', I_OR_IJ, 'd']],
      output=['ac', I_OR_IJ, 'bd'],
      how_to_apply='independent'
  )
  func = to_function(tt_einsum)
  return func(a, b)


def flat_inner(a, b):
  """Calculate inner product of given `TT-Tensors` or `TT-Matrices` wrapped with `WrappedTT`.
  
  :param a: first argument
  :type a: `WrappedTT`
  :param b: second argument
  :type b: `WrappedTT`
  :rerurn: the result of inner product
  :rtype: `WrappedTT`
  """
  tt_einsum = TTEinsum(
      inputs=[['a', I_OR_IJ, 'b'], ['c', I_OR_IJ, 'd'], ['a', 'c']],
      output=['b', 'd'],
      how_to_apply='cumulative'
  )
  func = to_function(tt_einsum)
  res = jnp.squeeze(func(a, b)[-1])
  return res


def matmul(a, b):
  """Calculate matrix multiplication of given `TT-Matrices` wrapped with `WrappedTT`.
  
  :param a: first argument
  :type a: `WrappedTT`
  :param b: second argument
  :type b: `WrappedTT`
  :rerurn: the result of inner product
  :rtype: `WrappedTT`
  """
  # TODO: support TT x dense matmul.
  tt_einsum = TTEinsum(
      inputs=[['a', 'ij', 'b'], ['c', 'jk', 'd']],
      output=['ac', 'ik', 'bd'],
      how_to_apply='independent'
  )
  func = to_function(tt_einsum)
  return func(a, b)


@tt_vmap()
def add(tt_a, tt_b):
  """Returns a `TT-object` corresponding to elementwise sum `tt_a + tt_b`.
  The shapes of `tt_a` and `tt_b` should coincide.
  Supports broadcasting, e.g. you can add a tensor train with
  batch size 7 and a tensor train with batch size 1:
  
  ``tt_batch.add(tt_single.batch_loc[np.newaxis])``
  
  where ``tt_single.batch_loc[np.newaxis]`` 
  creates a singleton batch dimension.

  :type tt_a: `TT-Tensor` or `TT-Matrix`
  :param tt_a: first argument
  :type tt_b: `TT-Tensor` or `TT-Matrix`
  :param tt_b: second argument
  :rtype: `TT-Tensor` or `TT-Matrix`
  :return: `tt_a + tt_b`
  :raises [ValueError]: if the arguments shapes do not coincide
  """
  tt_a = unwrap_tt(tt_a)
  tt_b = unwrap_tt(tt_b)

  if not are_shapes_equal(tt_a, tt_b):
    raise ValueError('Types of the arguments or their tensor '
                     'shapes are different, addition is not '
                     'available.')
  if not are_batches_broadcastable(tt_a, tt_b):
    raise ValueError('The batch sizes are different and not 1, '
                     'broadcasting is not available.')

  if tt_a.is_tt_matrix:
    tt_cores = _add_matrix_cores(tt_a, tt_b)
    return TTMatrix(tt_cores)
  else:
    tt_cores = _add_tensor_cores(tt_a, tt_b)
    return TT(tt_cores)


def _add_tensor_cores(tt_a, tt_b):
  """Internal function to be called from add for two TT-tensors.
  Does the actual assembling of the TT-cores to add two TT-tensors.
  """
  num_dims = tt_a.ndim
  shape = tt_a.shape
  a_ranks = tt_a.tt_ranks
  b_ranks = tt_b.tt_ranks
  tt_cores = []
  for core_idx in range(num_dims):
    a_core = tt_a.tt_cores[core_idx]
    b_core = tt_b.tt_cores[core_idx]
    if core_idx == 0:
      curr_core = jnp.concatenate((a_core, b_core), axis=2)
    elif core_idx == num_dims - 1:
      curr_core = jnp.concatenate((a_core, b_core), axis=0)
    else:
      upper_zeros = jnp.zeros((a_ranks[core_idx], shape[core_idx],
                              b_ranks[core_idx + 1]))
      lower_zeros = jnp.zeros((b_ranks[core_idx], shape[core_idx],
                              a_ranks[core_idx + 1]))
      upper = jnp.concatenate((a_core, upper_zeros), axis=2)
      lower = jnp.concatenate((lower_zeros, b_core), axis=2)
      curr_core = jnp.concatenate((upper, lower), axis=0)
    tt_cores.append(curr_core)
  return tt_cores


def _add_matrix_cores(tt_a, tt_b):
  """Internal function to be called from add for two TT-matrices.
  Does the actual assembling of the TT-cores to add two TT-matrices.
  """
  num_dims = tt_a.ndim
  shape = tt_a.raw_tensor_shape
  a_ranks = tt_a.tt_ranks
  b_ranks = tt_b.tt_ranks
  tt_cores = []
  for core_idx in range(num_dims):
    a_core = tt_a.tt_cores[core_idx]
    b_core = tt_b.tt_cores[core_idx]
    if core_idx == 0:
      curr_core = jnp.concatenate((a_core, b_core), axis=3)
    elif core_idx == num_dims - 1:
      curr_core = jnp.concatenate((a_core, b_core), axis=0)
    else:
      upper_zeros = jnp.zeros((a_ranks[core_idx], shape[0][core_idx],
                              shape[1][core_idx], b_ranks[core_idx + 1]))
      lower_zeros = jnp.zeros((b_ranks[core_idx], shape[0][core_idx],
                              shape[1][core_idx], a_ranks[core_idx + 1]))
      upper = jnp.concatenate((a_core, upper_zeros), axis=3)
      lower = jnp.concatenate((lower_zeros, b_core), axis=3)
      curr_core = jnp.concatenate((upper, lower), axis=0)
    tt_cores.append(curr_core)
  return tt_cores


def are_shapes_equal(tt_a, tt_b):
  """Returns the result of equality check of 2 tensors' shapes: 
  `True` if shapes are equal and `False` otherwise.
  The arguments should be both `TT-Tensors` or both `TT-Matrices`.
  The arguments should have the same tensor shape
  but their `TT-ranks` differ.

  :type tt_a: `TT-Tensor` or `TT-Matrix`
  :param tt_a: first argument to check
  :type tt_b: `TT-Tensor` or `TT-Matrix`
  :param tt_b: second argument to check
  :return: `tensor_check` - the result of shape check
  :rtype: bool
  """
  tensor_check = True
  if tt_a.is_tt_matrix != tt_b.is_tt_matrix:
    tensor_check = False
  if jnp.any(jnp.array(tt_a.raw_tensor_shape) != 
            jnp.array(tt_b.raw_tensor_shape)):
    tensor_check = False
  return tensor_check


def are_batches_broadcastable(tt_a, tt_b):
  """Returns the result of compatibility check of 2 tensors' batches: 
  `True` if batches are compatible and `False` otherwise.
  The batch sizes should be equal otherwise at least one of them 
  should equal to 1 for broadcasting to be available.
  
  :type tt_a: `TT-Tensor` or `TT-Matrix`
  :param tt_a: first argument to check
  :type tt_b: `TT-Tensor` or `TT-Matrix`
  :param tt_b: second argument to check
  :return: the result of broadcasting check
  :rtype: bool
  """
  if tt_a.num_batch_dims != tt_b.num_batch_dims:
    return False
  for a, b in zip(tt_a.batch_shape, tt_b.batch_shape):
    if a == 1 or b == 1 or a == b:
      pass
    else:
      return False
  return True


def multiply(a, b):
  """Calculate elementwise product of 2 `TT-Tensors` \ `TT-Matrices` or their product by scalar. Arguments could be wrapped by `WrappedTT` or not.
  
  :param a: first argument
  :type a: Union[float, TT-object]
  :param b: second argument
  :type b: Union[float, TT-object]
  :return: the result of elementwise product
  :rtype: `TT-object`
  """
  if not is_tt_object(a) or not is_tt_object(b):
    return multiply_by_scalar(a, b)
  else:
    return tt_tt_multiply(a, b)


def multiply_by_scalar(a, b):
  """Returns the result of multiplication so called `TT-object` 
  (`TTTensOrMat` or `WrappedTT`) by scalar. Takes 2 arguments 
  as input, one of which is `TT-object` and other is a scalar. 
  Does not depends on arguments order. 
  
  :return: the result of multiplication by scalar
  :rtype: `TTTensOrMat`
  """
  if is_tt_object(a):
    return _mul_by_scalar(a, b)
  else:
    return _mul_by_scalar(b, a)


@tt_vmap(1)
def _mul_by_scalar(tt, c):
  tt = unwrap_tt(tt)
  cores = list(tt.tt_cores)
  cores[0] = c * cores[0]
  if tt.is_tt_matrix:
    return TTMatrix(cores)
  else:
    return TT(cores)
